/* otackomer
   verzia 2.7
   31.10.2013
   
   
   ATtiny2313 - 20PU
   krystalovy oscilator 4 MHz (vydeleny na 1 MHz)
   LFUSE: 0xFD, HFUSE: 0x9F, EFUSE: 0xFF
   Flash 1596 B, SRAM 63 B, EEPROM 128 B
*/


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdint.h>

#define F_CPU 1000000UL
#include <util/delay.h>


#define LCD_DATA PORTB
#define RIADOK_MAX 8			// displej 16x1 znakov, takze sa k nemu pristupuje ako k 8x2

#define LCD_RL PORTD			// riadiaca linka pre LCD modul
#define RS 5
#define EN 4
#define PS 0					// podsvietenie

#define TL_PIN	 PIND			// tlacitka na porte D
#define TLAC1    3
#define TLAC2    2



void main(void) __attribute__ ((noreturn));
void napis(uint8_t adresa, const char* data);		// vypisanie retazca na displej
void prevod(uint16_t cislo);						// prevod 16b cisla na retazec
uint8_t nastavenie(void);							// nastavenie poctu listov a podsvietenia
uint8_t pocitanie(void);							// spocitanie stlaceni tlacitka


char displej_1[9], prazdny_str[] = "        ";		// retazec pre displej a medzery na vymazanie riadka
char *menu[] = {"Vrtula:","Svetlo ","Koniec "};

volatile uint16_t poc_stav, kon_stav;
volatile uint8_t impulzy, meranie = 1, pretecenia = 255;
// premenne nastavim tak aby sa po zapnuti hned vypocitalo 0ot/min

uint8_t __attribute__ ((section (".eeprom"))) EE_nastavenia = 4;		// predvolena hodnota su 2 listy a vypnute svetlo
#define PS_MASKA  0x01			// maska pre vycitanie podsvietenia z ulozeneho bajtu



void main(void) {
   DDRB = 0xFF;										// data pre displej
   DDRD = (1 << RS) | (1 << EN) | (1 << PS);		// vystupy: riadiace signaly pre displej, spinanie podsvietenia displeja
   ACSR = _BV(ACD);									// vypnem analogovy komparator kvoli znizeniu spotreby

   CLKPR = 0x80;									// nastavenie delicky systemovych hodin, pomer 4
   CLKPR = 0x02;



   //   *****   inicializacia displeja   *****

   LCD_RL = 1 << EN;								// nenastavujem jednotlive bity, je to pripustne, usetrim velkost kodu
   LCD_DATA = 0x30;									// 8 bitova zbernica
   LCD_RL = 0x00;
   _delay_ms(8);
   LCD_RL = 1 << EN;								// 8bit, kod prikazu sa nezmenil, len robim impulzy na vstupe EN
   LCD_RL = 0x00;
   _delay_us(250);
   LCD_RL = 1 << EN;								// 8bit
   LCD_RL = 0x00;
   _delay_us(250);

   LCD_RL = 1 << EN;
   LCD_DATA = 0x38;									// 2 riadky, 5x8 bodov
   LCD_RL = 0x00;
   _delay_us(60);

   LCD_RL = 1 << EN;
   LCD_DATA = 0x08;									// vypnut displej
   LCD_RL = 0x00;
   _delay_us(60);

   LCD_RL = 1 << EN;
   LCD_DATA = 0x01;									// vymazat displej
   LCD_RL = 0x00;
   _delay_ms(8);

   LCD_RL = 1 << EN;
   LCD_DATA = 0x06;									// posun doprava
   LCD_RL = 0x00;
   _delay_us(60);

   LCD_RL = 1 << EN;
   LCD_DATA = 0x0C;									// zapnut displej (kurzor a blikanie vypnute)
   LCD_RL = 0x00;
   _delay_us(60);

   LCD_DATA = 0x00;



   //   *****   nacitanie a zobrazenie ulozenych nastaveni   *****

   uint8_t adresa = 0, eedata;						// hladanie zacinam od adresy 0, hodnoty precitane z EEPROM
   uint8_t pocet_listov, nast;

   // hladanie volneho miesta v pamati
   // v pamati musi byt prvy bajt rozny od FF !

   do {												// hladam dokym neni precitany bajt 0xFF
      EEAR = adresa;
      EECR |= _BV(EERE);							// spustim citanie adresovaneho bajtu
      eedata = EEDR;								// precitam bajt
      adresa += 1;									// posuniem adresu o velkost jedneho bloku dat
   } while((eedata != 0xFF) && (adresa <= E2END));	// koncim hladanie ak som na konci EEPROM

   // urcenie adresy prveho volneho bajtu
   if(adresa <= E2END)								// ked je prvy volny bajt niekde pred koncom pamati
      adresa --;
   else {
      if(eedata != 0xFF)							// ked je pamat plna
         adresa = 0;
      else											// ked je volny posledny bajt
	     adresa = E2END;
   }


   EEAR = (adresa - 1) & E2END;						// precitam ulozene nastavenia
   EECR |= _BV(EERE);
   nast = EEDR;
   LCD_RL = nast & PS_MASKA;						// keby PS nebol PORTx0, tak bit z pamati nesedi s bitom portu
   pocet_listov = nast >> 1;
   prevod((uint16_t)pocet_listov);					// zobrazim ulozeny pocet listov
   napis(0x01,"Vrtula:");
   napis(0x41, displej_1 + 2);

   PORTD |= _BV(TLAC1);								// zvysovaci odpor pre tlacitko
   uint8_t i, j = 0;
   for(i=0; i<60; i++) {
      if(bit_is_clear(TL_PIN,TLAC1)) {				// ak je tlacitko stlacene,
         for(j=0; j<30; j++) {						// tak spustim cyklus pocitajuci 1,5s, pocas ktorych musi byt drzane
            _delay_ms(50);
            if(bit_is_set(TL_PIN,TLAC1))
			   break;
		 }
	  }
	  if(j >= 30) {									// ak bolo tlacitko drzane, bude sa nastavovat
		 //napis(0x40, "Nast>");
		 nast = nastavenie();						// zavolam funkciu na nastavovanie
		 pocet_listov = nast >> 1;					// hornych 7 bitov je informacia o vrtuli
	     break;										// po skonceni nastavovania ukoncim 3-sekundove cakanie
	  }
	  _delay_ms(50);								// ak tlacitko neni stlacene, tak len cakam na jeho stlacenie
   }

   if(nast != EEDR) {								// zapisem nastavenia do pamati, ale len vtedy ked sa zmenili
      if(adresa == 0) {								// ked je pamat plna, treba ju vymazat
	     for( ; adresa <= E2END; adresa ++) {		// prejdem celu pamat, hodnotu adresy mozem menit, potom ju vratim
	  	    EEAR = adresa;
	  	    EECR = _BV(EEPM0) | _BV(EEMPE);			// mazanie, povolene prerusenie | _BV(EERIE)
	  	    EECR |= _BV(EEPE);						// spustenie mazania adresovanej bunky
	  	    loop_until_bit_is_clear(EECR, EEPE);	// cakanie na dokoncenie operacie
   		 }
   		 adresa = 0;								// povodny stav pred mazanim
	  }
	  // zapis nastaveni do pamati
	  EEAR = adresa;
	  EEDR = nast;
      EECR = _BV(EEPM1) | _BV(EEMPE);				// rezim iba zapis
      EECR |= _BV(EEPE);
   }


   //   *****   napisanie napisu "ot/min"   *****

   napis(0x00, prazdny_str);
   napis(0x40, prazdny_str);
   napis(0x41, "ot/min");						// napisanie jednotiek na 10. miesto


   //   *****   nastavenie a zapnutie casovacov   *****

   OCR0A = 244;									// casovac0 sluzi na taktovanie programu, dava prerusenie kazdych ~250 ms
   TCCR0A = _BV(WGM01) & (~_BV(WGM00));			// nastaveny rezim CTC
   TCCR0B = _BV(CS02) | _BV(CS00);				// taktovanie systemovymi hodinami, deliaci pomer 1024
   TCCR1B = _BV(CS10);							// zapnuty casovac na systemove hodiny, zachytenie vstupu pri dobeznej hrane
   TIMSK = _BV(OCIE0A);							// prerusenie na taktovanie programu, ostatne sa spustia v jeho obsluhe

   uint16_t otacky;
   uint8_t necinnost = 0;						// ak dlho nebudu prichdzat impulzy, tak sa vypina podsvietenie displeja

   sei();
   while(1) {
	 //   ***** vypocet   *****
     uint32_t T;								// perioda
     if(meranie) {
	    cli();									// pred vyhodnocovanim vypnem prerusenia, takze zabranim zmenam hodnot
	 	if(pretecenia > 15) {					// pri dlhej dobe bez impulzu, cca 1 s
   	 	   otacky = 0;							// povazujem otacky za nulove
   	 	   prevod(otacky);
   	 	   napis(0x03, displej_1);
   	 	   pretecenia = 0;						// pretecenia treba vycistit
		   meranie = 0;							// tym padom povazujem meranie za dokoncene
		   TIMSK = _BV(OCIE0A);					// v case medzi meraniami niesu prerusenia casovaca1 potrebne
   	 	   impulzy = 0;
   	 	   necinnost ++;
   	 	}
     	else {
   	       if(impulzy == 5) {
			  T = pretecenia * 65536U + kon_stav - poc_stav;
			  T *= pocet_listov;
   	  	   	  otacky = 240000000UL / T;			// vysledok v [ot/min]
   	  	   	  prevod(otacky);
   	  	   	  napis(0x03, displej_1);
   	  	   	  meranie = 0;						// meranie dokoncene, cakam kym prerusenie OC0A da povel na nove meranie
   	  	   	  TIMSK = _BV(OCIE0A);				// v case medzi meraniami niesu prerusenia casovaca1 potrebne
			  impulzy = 0;
   	  	   	  necinnost = 0;
  		   }
     	}
     	if(necinnost >= 19) {					// po 20 s kedy su nulove otacky sa vypne podsvietenie
     	   LCD_RL &= ~_BV(PS);
     	   necinnost = 19;						// aby pocitadlo necinnosti nepretieklo
	    }
  	    else									// inac zostava podsvietenie v nastavenom stave
  	       LCD_RL |= nast & PS_MASKA;
	    sei();									// po skonceni vyhodnocovania znova povolim prerusenia
	 }

	 MCUCR = _BV(SE);
	 sleep_cpu();								// rezim Idle
	 MCUCR = 0x00;
   } // koniec hlavneho cyklu

} // koniec main



//   ***** prerusenia *****
//   **********************

ISR(TIMER1_CAPT_vect) {
   if(meranie)
      impulzy ++;

   if(impulzy == 1) {					// prvy impulz zacina meranie
     poc_stav = ICR1;					// stav casovaca na pociatku merania
     pretecenia = 0;					// na zaciatku merania treba vynulovat
   }
   if(impulzy == 5) {					// meram trvanie styroch impuzlov, takze 5. ukoncuje meranie
     kon_stav = ICR1;					// stav casovaca na konci merania
   }
}


ISR(TIMER1_OVF_vect) {
   if(meranie)							// pretecenia pocitam len pocas merania
     pretecenia ++;
}


ISR(TIMER0_COMPA_vect) {				// dava signal na spustenie merania cez premennu meranie
   static uint8_t prerusenia;			// chcem taktovanie 500ms, casovac dava prerusenie kazdych 250ms, takze ich treba pocitat

   if(prerusenia == 0 && !meranie) {	// spustenie noveho merania, len ak je predchadzajuce dokoncene
      meranie = 1;
      TIFR = _BV(TOV1) | _BV(ICF1);		// prerusenia casovaca1 znova povolim, ale najprv treba vymazat priznaky
      TIMSK = _BV(ICIE1) | _BV(TOIE1) | _BV(OCIE0A);
   }
   prerusenia ^= 0x01;					// premenna sa sprava ako 1 bitovy citac preruseni
}



//   *****   nastavovanie   *****
//   ****************************
uint8_t nastavenie(void) {
   uint8_t  i, m_i = 0;		 						// menu_index
   uint8_t vystup = EEDR;							// vyst funkcie, inicializuje sa aktualnym stavom

   napis(0x40, prazdny_str);						// prva polozka je vrtula, to uz zobrazene je, len vymazem cislo
   loop_until_bit_is_set(TL_PIN, TLAC1);			// pred pokracovanim musim pockat na pustenie tlac
   _delay_ms(50);									// cakanie na dokmitanie tlacitka


   while(1) {
	  if(bit_is_clear(TL_PIN,TLAC1)) {				// pri stlacenom tlacitku
		 for(i=0; i<30; i++) {
			if(bit_is_set(TL_PIN, TLAC1)) {
			   if(++m_i == sizeof(menu)/sizeof(char *))	// inkrementacia s vytvorenim pretecenia
			      m_i = 0;
			   napis(0x01, menu[m_i]);				// zobrazi sa dalsia polozka menu
			   break;
			}
			_delay_ms(50);
		 }
		 if(i == 30) {								// pri podrzani tlacitka sa nastavenie ukonci
		    if(m_i == 0) {							// ak bolo tlacitko podrzane na polozke Vrtula, idu sa pocitat listy
   			   vystup = (pocitanie() << 1) | (vystup & PS_MASKA);
			   napis(0x40, prazdny_str);			// vymazem cislo po nastavovani
			   //loop_until_bit_is_set(TL_PIN, TLAC1);	// aby sa zobrazenie nepreplo hned na dalsiu pozlozku
  	        }
			if(m_i == 1) {							// ak bolo tlacitko podrzane na polozke svetlo
			   LCD_RL ^= _BV(PS);					// negujem stav podsvietenia
			   vystup ^= _BV(PS);					// negujem ho aj v premennej
			   loop_until_bit_is_set(TL_PIN, TLAC1);	// aby sa zobrazenie nepreplo hned na dalsiu pozlozku
			}
		    if(m_i == 2)							// ak bolo podrzane na polozke Koniec
		       break;								// skoncim nekonecny cyklus cize aj funkciu nastavenie
		 }
	  }
	  _delay_ms(50);
   }
   return vystup;
}


//   *****   spocitanie stlaceni tlacitka   *****
//   ********************************************

uint8_t pocitanie(void) {
   uint8_t i = 0, listy = 1;

   prevod((uint16_t)listy);						// zobrazim zaciatocny pocet listov, sluzi aj ako signal, ze sa moze pustit tlacitko
   napis(0x41, displej_1 + 2);
   loop_until_bit_is_set(TL_PIN, TLAC1);		// cakanie na pustenie tlacitka pri zacati nastavovania
   _delay_ms(50);								// kvoli zakmitom tlacitka

   while(1) {
      if(i < 30) {								// ak uplynulo 1,5s nespusti sa pocitanie znova
         if(bit_is_clear(TL_PIN, TLAC1)) {		// nastavit na aku nozicku bude pripojene tlacitko
            for(i=0; i<30; i++) {				// pri drzani tlacitka pocita 30*50ms=1,5s
      	       _delay_ms(50);
      	       if(bit_is_set(TL_PIN, TLAC1)) {	// ak sa pusti tlacitko, pripocita list a ukonci cyklus pocitajuci 1,5s a znova sa spusti kontrola ci je stlacene tlacitko
      		      listy++;
				  if(listy > 126)				// max povoleny pocet => pretecenie
				     listy = 1;
      		      prevod((uint16_t) listy);		// zobrazi sa aktualny pocet listov
      		      napis(0x41, displej_1 + 2);
      		      break;
  		       }
       	    }
         }
      }
      if(i >= 30)								// ak bolo tlacitko podrzane, tak sa konci pocitanie
         break;
	  _delay_ms(50);							// cyklus sa opakuje s periodou 50 ms
   }

   return listy;
}



//   *****   prevod 16b cisla na retazec   *****
//   *******************************************

void prevod(uint16_t cislo) {
/* funkcia prevedie 16b bezznamienkove cislo na retazec, zapisuje do pola displej_1, od jeho zaciatku
   uvodne nuly nahradzuje medzerami, zapisuje vzdy 5 cislic a ukoncovaci znak (spolu 6B)
*/

   uint8_t i = 0, znak;

   // prva cislica
   znak = 0x30;
   if(cislo > 9999) {
      znak = cislo / 10000;
      cislo %= 10000;
      znak |= 0x30;
   }
   else
      znak = 0x20;
   displej_1[i] = znak;

   // druha cislica
   if((znak != 0x20) && (cislo < 1000))
      znak = 0x30;
   else {
      if(cislo > 999) {
         znak = cislo / 1000;
      	 cislo %= 1000;
      	 znak |= 0x30;
      }
      else
         znak = 0x20;
   }
   i ++;
   displej_1[i] = znak;

   // tretia cislica
   if((znak != 0x20) && (cislo < 100))
      znak = 0x30;
   else {
      if(cislo > 99) {
         znak = cislo / 100;
      	 cislo %= 100;
      	 znak |= 0x30;
      }
      else
         znak = 0x20;
   }
   i ++;
   displej_1[i] = znak;

   // stvrta cislica
   if((znak != 0x20) && (cislo < 10))
      znak = 0x30;
   else {
      if(cislo > 9) {
         znak = cislo / 10;
      	 cislo %= 10;
      	 znak |= 0x30;
      }
      else
         znak = 0x20;
   }
   i ++;
   displej_1[i] = znak;

   // piata cislica
   i ++;
   znak = 0x30 | cislo;
   displej_1[i] = znak;

   // ukoncovaci znak
   i ++;
   displej_1[i] = '\0';
}



//   *****   vypisovacia funkcia   *****
//   ***********************************

void napis(uint8_t adresa, const char *data) {
/* funkcia posle retazec z adresy data na displej, zacne na znaku so zadanou adresou
   funkcia vypisuje pokial nepride k ukoncovaciemu znaku \0, ale najviacej RIADOK_MAX znakov
   to ci sa retazec do riadka na displeji vojde musi zvazit volajuci, funkcia to neriesi
   komunikacia s LCD modulom je 8 bitova, urcene pre 1 MHz systemove hodiny, pre vyssiu f treba zaistit spravne casovanie
*/

   //   ***  nastavenie adresy LCD modulu kde sa ma pisat text   ***
   LCD_RL &= ~_BV(RS);						// nulujem RS
   LCD_RL |= _BV(EN);						// nastavim E
   LCD_DATA = 0x80 | adresa;				// na displej poslem prikaz zmeny adresy s novou adresou
   LCD_RL &= ~_BV(EN);						// dobeznou hranou na E potvrdim prikaz
   _delay_us(60);

   //   ***  odosielanie znakov do LCD modulu  ***
   LCD_RL |= _BV(RS);						// nastavim RS kedze idem posielat data
   uint8_t i;
   for(i=0; i < RIADOK_MAX && *data != '\0'; i++) {
      /*if(*data == '\0')						// ak je aktualny znak ukoncovaci, tak koncim vypisovanie
         break;*/
	  LCD_RL |= _BV(EN);					// nastavim E
      LCD_DATA = *data;
      LCD_RL &= ~_BV(EN);					// dobeznou hranou na E potvrdim data
      data ++;								// posuniem pointer na dalsi znak
      _delay_us(60);
   }
   // return i;								// v pripade, ze by som chcel vracat pocet vypisanych(odoslanych) znakov
}
